#include <windows.h>
#include "directory.h"
#include "file.h"
#include "vm.h"

static HANDLE dirs [SLATE_FILES_MAXIMUM];

void initDirectoryModule ()
{
	memset(dirs, 0, sizeof(dirs));
}

int
allocateDirectory (void)
{
	int dir;

	for (dir = 0; dir < SLATE_FILES_MAXIMUM; ++ dir)
	{
		if (dirs [dir] == 0)
			return dir;
	}

	return -1;
}

void
closeDirectory (int dir)
{
	if (dirs [dir] != 0)
	{
		CloseHandle(dirs [dir]);

		dirs [dir] = 0;
	}
}

int
openDirectory (struct ByteArray * name, int flags)
{
	char nameString[SLATE_FILE_NAME_LENGTH];
	size_t nameLength = extractCString (name, nameString, sizeof(nameString));
	DWORD rwMode = 0;
	DWORD openMode = 0;
	int dir;

	if (nameLength <= 0 || nameLength > MAX_PATH)
		return -1;
		
	dir = allocateDirectory ();
	if (dir < 0)
		return -1;

	if (flags & SF_WRITE)
		rwMode |= GENERIC_WRITE;
	if (flags & SF_READ)
		rwMode |= GENERIC_READ;

	dirs [dir] = CreateFile(nameString, rwMode, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

	if (dirs [dir] == INVALID_HANDLE_VALUE)
	{
		// nothing exists with this name
		if (flags & SF_CREATE)
		{
			if (CreateDirectory(nameString, NULL))
			{
				dirs [dir] = CreateFile(nameString, rwMode, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
				if (dirs [dir] == INVALID_HANDLE_VALUE)
					goto failReturn;
			}
		}
		else
		{
			goto failReturn;
		}
	}
	else
	{
		if (!(GetFileAttributes(dirs [dir]) & FILE_ATTRIBUTE_DIRECTORY))
			goto failReturn;
	}

	return dir;

failReturn:
	if (dirs [dir] != 0)
	{
		if (dirs [dir] > 0)
			CloseHandle (dirs [dir]);

		dirs [dir] = 0;
	}
	return -1;
}

struct ByteArray *
getCurrentDirectory ()
{
	struct ByteArray * result;
	char pathString [SLATE_PATH_NAME_LENGTH];

	DWORD pathLength = GetCurrentDirectory (sizeof(pathString), pathString);

	if (pathLength < 0)
		return Nil;

	// TODO write a pidgin primitive for this
	result = PSObjectHeap_newByteArray_sized_ (CurrentMemory, ByteArrayProto, pathLength);
	//initializeByteArray (pathString, pathLength, result);

	return result;
}

int
setCurrentDirectory (struct ByteArray * path)
{
	char pathString[SLATE_PATH_NAME_LENGTH];
	ssize_t pathLength = extractCString (path, pathString, sizeof(pathString));

	if (pathLength < 0)
		return -1;

	return SetCurrentDirectory (pathString);
}
